using System;
using System.Collections.Generic;
using System.Text;

// 
// NOTE: Listing B-19 in the book is illustrative only
//

namespace DevLeap.Linq.CSharp.LambdaExpressions {

    class AggLambdaExpression {
        public List<int> Values;

        /// <summary>
        /// Listing B-16
        /// 
        /// Lambda expressions instead of anonymous methods
        ///   - delegate keyword disappears 
        ///   - add "=>" between parameters list and code
        /// </summary>
        public static void Listing2_16() {
            AggLambdaExpression l = new AggLambdaExpression();
            int sum, min;
            sum = Aggregate(
                      l.Values,
                      (int a, int b) => { return a + b; }
                   );
            min = Aggregate(
                      l.Values,
                      (int a, int b) => { return (a < b) ? a : b; }
                   );
            Console.WriteLine("Sum = {0}", sum);
            Console.WriteLine("Min = {0}", min);
        }

        /// <summary>
        /// Clarify this: lambda expression definition with a generic type
        ///   - delegate keyword disappears 
        ///   - add "=>" between parameters list and code
        /// </summary>
        delegate int Funcint<T>(T a, T b);
        public static void Demo01T<T>() {
            Funcint<T> x = (a, b) => { return (a.ToString().Length + b.ToString().Length); };
            Console.WriteLine("Length sum = {0}", x);
        }

        /// <summary>
        /// Listing B-17
        /// 
        /// Lambda expressions without types
        ///   - parameter type is not defined
        ///   - lambda expression resolution is deferred to lambda expression use
        /// </summary>
        public static void Listing2_17() {
            AggLambdaExpression l = new AggLambdaExpression();
            int sum, min;
            sum = Aggregate(
                      l.Values,
                      (a, b) => { return a + b; }
                   );
            min = Aggregate(
                      l.Values,
                      (a, b) => { return (a < b) ? a : b; }
                   );
            Console.WriteLine("Sum = {0}", sum);
            Console.WriteLine("Min = {0}", min);
        }

        /// <summary>
        /// Listing B-18
        /// 
        /// Lambda expressions without types
        ///   - no code block, only an expression is allowed after "=>"
        /// </summary>
        public static void Listing2_18() {
            AggLambdaExpression l = new AggLambdaExpression();
            int sum, min;
            sum = Aggregate(
                      l.Values,
                      (a, b) => a + b
                   );
            min = Aggregate(
                      l.Values,
                      (a, b) => (a < b) ? a : b
                   );
            Console.WriteLine("Sum = {0}", sum);
            Console.WriteLine("Min = {0}", min);
        }

        /// <summary>
        /// Lambda expressions accessing external data
        ///   - lambda expression parameter is important to understand identifiers inside lambda expression
        /// </summary>
        public static void Demo04() {
            AggLambdaExpression l = new AggLambdaExpression();
            int sum = 0;
            int min = 0;
            bool firstLoop = true;
            sum = AggregateSingle(
                      l.Values,
                      x => sum += x  // x specify what are the params and what are reference to existing identifiers...
                   );
            min = AggregateSingle(
                      l.Values,
                      x => {
                          if (firstLoop) {
                              min = x;
                              firstLoop = false;
                          }
                          else {
                              min = (x < min) ? x : min;
                          }
                          return min;
                      }
                   );

            Console.WriteLine("Sum = {0}", sum);
            Console.WriteLine("Min = {0}", min);
        }

        public AggLambdaExpression() {
            Values = new List<int>();
            Values.Add(10);
            Values.Add(20);
            Values.Add(30);
            Values.Add(40);
            Values.Add(50);
        }

        delegate T Func<T>(T a, T b);
        static T Aggregate<T>(List<T> l, Func<T> f) {
            T result = default(T);
            bool firstLoop = true;
            foreach (T value in l) {
                if (firstLoop) {
                    result = value;
                    firstLoop = false;
                }
                else {
                    result = f(result, value);
                }
            }
            return result;
        }

        delegate T FuncSingle<T>(T a);
        static T AggregateSingle<T>(List<T> l, FuncSingle<T> f) {
            T result = default(T);
            foreach (T value in l) {
                result = f(value);
            }
            return result;
        }

        delegate T FuncNoPar<T>();
        static void Fill<T>(List<T> l, FuncNoPar<T> f) {
            for (int i = 0; i < l.Count; i++) {
                l[i] = f();
            }
        }
    }

}