﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data.EntityClient;
using System.Data.Common;
using System.Data;
using NorthwindModel;
using System.Data.Objects;
using System.Transactions;
using System.Data.Objects.DataClasses;
using System.Data.SqlClient;
using System.Configuration;

namespace DevLeap.EntityFramework
{
    class Program
    {
        static void Main(string[] args)
        {
            // ************************************************************************************
            // Chapter 8 - Code Samples - LINQ to Entities
            // ************************************************************************************

            LinqToEntitiesQueryWithFilter();
            LinqToEntitiesEquivalentEntityQuery();
            LinqToEntitiesWithMultipleFiltersOnRelationships();
            LinqToEntitiesExplicitLoadOfRelationships();
            LinqToEntitiesIncludeRelationships();
            LinqToEntitiesUnionSample();
            LinqToEntitiesQueryWithCalculations();
            LinqToEntitiesDataModification();
            LinqToEntitiesServerSideEvaluation();
            LinqToEntitiesClientSideEvaluation();
            LinqToEntitiesObjectIdentityAndMergeOption();
            LinqToEntitiesShowInnerTypeOfQueryVariable();
            LinqToEntitiesCompiledQuery();

            // ************************************************************************************
            // Appendix A - Code Samples - ADO.NET Entity Framework
            // ************************************************************************************

            LegacySqlClient();
            LegacyProviderFactories();
            EntityCommandExplicit();
            EntityQueryExplicit();
            EntityQueryTypedExplicit();
            EntityQueryTypedWithFilterExplicit();
            EntityQueryTypedWithParametricFilterExplicit();
            EntityQueryTypedWithParametricFilterOverInnerCollectionExplicit();
            EntityQueryTypedWithNavigateExplicit();
            EntityQueryTypedWithNavigateAndCachingExplicit();
            LinqToEntitiesFilteredDistinct();
            EntityModificationAndPersistence();
            EntityModificationWithTransactionSample();
        }

        #region Chapter 8 - Code Samples - LINQ to Entities

        // Listing 8-1

        private static void LinqToEntitiesQueryWithFilter()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {

                var italianCustomers =
                    from c in northwind.Customers
                    where c.Country == "Italy"
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing 8-2

        private static void LinqToEntitiesEquivalentEntityQuery()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomersObjectQuery =
                    new ObjectQuery<Customers>(
                        @"SELECT VALUE Customers FROM " +
                        @"NorthwindEntities.Customers " +
                        @"WHERE Customers.Country = 'Italy'",
                        northwind);

                northwind.Connection.Open();
                Console.WriteLine(italianCustomersObjectQuery.ToTraceString());

                foreach (var c in italianCustomersObjectQuery)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing 8-3

        private static void LinqToEntitiesWithMultipleFiltersOnRelationships()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomerWithAtLeastTenOrders =
                    from c in northwind.Customers
                    where c.Country == "Italy"
                        && c.Orders.Count >= 10
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomerWithAtLeastTenOrders).ToTraceString());

                foreach (var c in italianCustomerWithAtLeastTenOrders)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing 8-4

        private static void LinqToEntitiesExplicitLoadOfRelationships()
        {

            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomers =
                    from c in northwind.Customers
                    where c.Country == "Italy"
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());

                    // Only for customers with PostalCode 42100
                    if (c.PostalCode == "42100")
                    {
                        // Load their Orders if they have not yet been loaded
                        if (!c.Orders.IsLoaded)
                        {
                            c.Orders.Load();
                        }

                        // Browse orders loaded
                        foreach (Orders o in c.Orders)
                        {
                            Console.WriteLine("\tOrderID: {0} - OrderDate: {1}", o.OrderID, o.OrderDate);
                        }
                    }
                }
            }
        }

        // Listing 8-5

        private static void LinqToEntitiesIncludeRelationships()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomers =
                    from c in northwind.Customers.Include("Orders")
                    where c.Country == "Italy"
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());

                    foreach (Orders o in c.Orders)
                    {
                        Console.WriteLine("\tOrderID: {0} - OrderDate: {1}", o.OrderID, o.OrderDate);
                    }
                }
            }
        }

        // Listing 8-6

        private static void LinqToEntitiesUnionSample()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomers =
                    new ObjectQuery<Customers>("SELECT VALUE Customers FROM " +
                    "NorthwindEntities.Customers WHERE Customers.Country = 'Italy'",
                    northwind);

                var germanCustomers =
                    new ObjectQuery<Customers>("SELECT VALUE Customers FROM " +
                    "NorthwindEntities.Customers WHERE Customers.Country = 'Germany'",
                    northwind);

                northwind.Connection.Open();
                Console.WriteLine(italianCustomers.ToTraceString());
                Console.WriteLine(germanCustomers.ToTraceString());

                foreach (var c in italianCustomers.UnionAll(germanCustomers))
                {
                    Console.WriteLine(c.Display());
                }
            }
        }
        // Listing 8-7

        private static void LinqToEntitiesQueryWithCalculations()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomersWithOrdersStats =
                    from c in northwind.Customers
                    where c.Country == "Italy"
                    select new
                    {
                        c.CustomerID,
                        c.ContactName,
                        c.CompanyName,
                        OrdersTotalCount = c.Orders.Count
                    };

                foreach (var c in italianCustomersWithOrdersStats)
                {
                    Console.WriteLine(c);
                }
            }
        }

        // Listing 8-8

        private static void LinqToEntitiesDataModification()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                DateTime today = DateTime.Now;

                var ordersToShip =
                    from o in northwind.Orders
                    where o.RequiredDate <= today
                       && o.ShippedDate == null
                    select o;

                foreach (var o in ordersToShip)
                {
                    o.ShippedDate = today;
                }

                using (TransactionScope scope = new TransactionScope())
                {
                    northwind.SaveChanges();
                    scope.Complete();
                }
            }
        }

        private static void LinqToEntitiesServerSideEvaluation()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var italianCustomers =
                    from c in northwind.Customers
                    where c.Country == "Italy"
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing 8-9

        private static void LinqToEntitiesClientSideEvaluation()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                String country = "Italy";

                var italianCustomers =
                    from c in northwind.Customers
                    where c.Country == country
                    select c;

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing 8-10

        private static void LinqToEntitiesObjectIdentityAndMergeOption()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                String country = "Germany";

                var germanCustomers =
                    (from c in northwind.Customers
                     where c.Country == country
                     select c).Take(4);

                //northwind.Connection.Open();
                //Console.WriteLine(((ObjectQuery<Customers>)germanCustomers).ToTraceString());

                Console.WriteLine("=> First query output");
                foreach (var c in germanCustomers)
                {
                    Console.WriteLine(c.Display());
                }

                Customers firstGermanCustomer = germanCustomers.First();
                firstGermanCustomer.ContactName += " - Modified!";

                var secondQueryForGermanCustomers =
                    (from c in northwind.Customers
                     where c.Country == country
                     select c).Take(4);

                // ((ObjectQuery<Customers>)secondQueryForGermanCustomers).MergeOption = MergeOption.OverwriteChanges;

                Console.WriteLine("=> Second query output");
                foreach (var c in secondQueryForGermanCustomers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }


        private static void LinqToEntitiesShowInnerTypeOfQueryVariable()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                String country = "Germany";

                var germanCustomersProjection =
                    from c in northwind.Customers
                    where c.Country == country
                    select new
                    {
                        c.CustomerID,
                        c.ContactName,
                        c.Country,
                    };

                Console.WriteLine(germanCustomersProjection.GetType());

                foreach (var c in germanCustomersProjection)
                {
                    Console.WriteLine(c);
                }
            }
        }

        private static void LinqToEntitiesCompiledQuery()
        {
            using (var northwind = new NorthwindModel.NorthwindEntities())
            {
                var compiledQueryForCustomersByCountry =
                    CompiledQuery.Compile<NorthwindEntities, String, IQueryable<Customers>>(
                    (context, country) => from c in context.Customers
                                          where c.Country == country
                                          select c);

                Console.WriteLine("=> German customers");
                var germanCustomers = compiledQueryForCustomersByCountry(northwind, "Germany");

                northwind.Connection.Open();
                Console.WriteLine(((ObjectQuery<Customers>)germanCustomers).ToTraceString());

                foreach (var c in germanCustomers)
                {
                    Console.WriteLine(c.Display());
                }

                Console.WriteLine("=> Italian customers");
                var italianCustomers = compiledQueryForCustomersByCountry(northwind, "Italy");

                Console.WriteLine(((ObjectQuery<Customers>)italianCustomers).ToTraceString());

                foreach (var c in italianCustomers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        #endregion

        #region Appendix A - Code Samples - ADO.NET Entity Framework

        // Listing A-1

        private static void LegacySqlClient()
        {
            SqlConnection cn = new SqlConnection(ConfigurationManager.ConnectionStrings["Northwind"].ConnectionString);
            SqlCommand cmd = new SqlCommand("SELECT c.CustomerID, c.CompanyName, " +
            "c.ContactName, o.OrderID, o.OrderDate FROM Customers AS c INNER JOIN " +
            "Orders AS o ON c.CustomerID = o.CustomerID", cn);
            using (cn)
            {
                cn.Open();
                using (SqlDataReader dr = cmd.ExecuteReader(CommandBehavior.CloseConnection))
                {
                    // Use the reader
                    while (dr.Read())
                    {
                        Console.WriteLine(dr[0]);

                        // In Windows Forms data-binnding it should be
                        // gridCustomersOrders.DataSource = dr;
                        // gridCustomersOrders.DataBind();
                    }
                }
            }
        }

        // Listing A-2

        private static void LegacyProviderFactories()
        {
            DbProviderFactory dbFactory = DbProviderFactories.GetFactory(ConfigurationManager.ConnectionStrings["Northwind"].ProviderName);

            DbConnection cn = dbFactory.CreateConnection();
            cn.ConnectionString = ConfigurationManager.ConnectionStrings["Northwind"].ConnectionString;

            DbCommand cmd = cn.CreateCommand();
            cmd.CommandText = "SELECT c.CustomerID, c.CompanyName, c.ContactName, " +
                "o.OrderID, o.OrderDate FROM Customers AS c INNER JOIN Orders AS o " +
                "ON c.CustomerID = o.CustomerID";

            using (cn)
            {
                cn.Open();
                using (DbDataReader dr = cmd.ExecuteReader(CommandBehavior.CloseConnection))
                {
                    // Use the reader
                    while (dr.Read())
                    {
                        Console.WriteLine(dr[0]);

                        // In Windows Forms data-binnding it should be
                        // gridCustomersOrders.DataSource = dr;
                        // gridCustomersOrders.DataBind();
                    }
                }
            }
        }

        // Listing A-8

        private static void EntityCommandExplicit()
        {
            using (EntityConnection cn = new EntityConnection(System.Configuration.ConfigurationManager.ConnectionStrings["NorthwindEntities"].ConnectionString))
            {
                EntityCommand cmd = new EntityCommand(
                    "SELECT c FROM NorthwindEntities.Customers AS c", cn);

                cn.Open();
                using (DbDataReader dr =
                           cmd.ExecuteReader(CommandBehavior.SequentialAccess))
                {
                    while (dr.Read())
                    {
                        Console.WriteLine(((DbDataRecord)dr.GetValue(0))[0]);
                    }
                }
            }
        }

        // Listing A-11

        private static void EntityQueryExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {

                var customers = db.CreateQuery<DbDataRecord>(
                    "SELECT c FROM NorthwindEntities.Customers AS c");

                foreach (var c in customers)
                {
                    Console.WriteLine(((Customers)c[0]).Display());
                }
            }
        }

        // Listing A-12

        private static void EntityQueryTypedExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {

                var customers = db.CreateQuery<Customers>(
                    "SELECT VALUE c FROM NorthwindEntities.Customers AS c");

                foreach (var c in customers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing A-13

        private static void EntityQueryTypedWithFilterExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {

                var customers = db.CreateQuery<Customers>(
                    "SELECT VALUE c FROM NorthwindEntities.Customers AS c " +
                    "WHERE c.Country = 'Italy'");

                foreach (var c in customers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing A-14

        private static void EntityQueryTypedWithParametricFilterExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {
                String country = "Italy";

                var customers = db.CreateQuery<Customers>(
                    "SELECT VALUE c FROM NorthwindEntities.Customers AS c " +
                    "WHERE c.Country = @Country");

                using (db.Connection)
                {
                    db.Connection.Open();
                    customers.Parameters.Add(new ObjectParameter("Country", country));

                    Console.WriteLine(customers.ToTraceString());

                    foreach (Customers c in customers)
                    {
                        Console.WriteLine(c.Display());
                    }
                }
            }
        }

        // Listing A-15

        private static void EntityQueryTypedWithParametricFilterOverInnerCollectionExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {
                String country = "Italy";

                ObjectQuery<Orders> orders = db.CreateQuery<Orders>(
                    "SELECT VALUE o FROM NorthwindEntities.Orders AS o " +
                    "WHERE o.Customers.Country = @Country");
                orders.Parameters.Add(new ObjectParameter("Country", country));

                foreach (Orders o in orders)
                {
                    Console.WriteLine(o.Display());
                }
            }
        }

        // Listing A-16

        private static void EntityQueryTypedWithNavigateExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {
                var customers = db.CreateQuery<Customers>(
                    "SELECT VALUE c FROM NorthwindEntities.Customers AS c " +
                    "WHERE EXISTS( " +
                    "SELECT VALUE o " +
                    "FROM NAVIGATE(c, NorthwindModel.FK_Orders_Customers) " +
                    "AS o " +
                    "WHERE o.OrderDate > @Date)");
                customers.Parameters.Add(new ObjectParameter("Date",
                    DateTime.Now.AddYears(-10)));

                foreach (var c in customers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing A-17

        private static void EntityQueryTypedWithNavigateAndCachingExplicit()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {
                var customers = db.CreateQuery<Customers>(
                    "SELECT VALUE c FROM NorthwindEntities.Customers AS c " +
                    "WHERE EXISTS( " +
                    " SELECT VALUE o " +
                    " FROM NAVIGATE(c, NorthwindModel.FK_Orders_Customers)" +
                    " AS o " +
                    " WHERE o.OrderDate > @Date)");
                customers.Parameters.Add(new ObjectParameter("Date",
                    DateTime.Now.AddYears(-10)));

                customers.EnablePlanCaching = true;

                foreach (var c in customers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing A-18

        private static void LinqToEntitiesFilteredDistinct()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {

                DateTime referenceDate = DateTime.Now.AddYears(-10);
                var customers = (from o in db.Orders
                                 where o.OrderDate > referenceDate
                                 select o.Customers).Distinct();

                foreach (var c in customers)
                {
                    Console.WriteLine(c.Display());
                }
            }
        }

        // Listing A-21

        private static void EntityModificationWithTransactionSample()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {
                using (TransactionScope scope = new TransactionScope())
                {

                    var customer = (from c in db.Customers
                                    where c.CustomerID == "ALFKI"
                                    select c).AsEnumerable().First();

                    customer.ContactName = "Paolo Pialorsi";

                    db.SaveChanges();
                    scope.Complete();
                }
            }
        }

        // Listing A-19 and A-20

        private static void EntityModificationAndPersistence()
        {
            using (NorthwindEntities db = new NorthwindEntities())
            {

                var customer = (from c in db.Customers
                                where c.CustomerID == "ALFKI"
                                select c).AsEnumerable().First();

                Console.WriteLine("Customer ContactName: {0}", customer.ContactName);
                Console.WriteLine("Customer state: {0}", customer.EntityState);

                customer.ContactName = "Maria Anders - Changed";

                Console.WriteLine("Customer ContactName: {0}", customer.ContactName);
                Console.WriteLine("Customer state: {0}", customer.EntityState);

                db.SaveChanges();

                Console.WriteLine("Customer state: {0}", customer.EntityState);
            }
        }

        #endregion
    }

    static class NorthwindExtender
    {
        public static String Display(this Customers c)
        {
            return (String.Format("{0} - {1}",
                c.CustomerID, c.ContactName));
        }

        public static String Display(this Orders o)
        {
            return (String.Format("{0} - {1}",
                o.OrderID, o.ShipName));
        }
    }
}
