﻿using System;
using System.Reflection;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;

namespace DevLeap.Linq.ExpressionTrees.DynamicFilter {
    // 
    // Listing 11-28
    // 
    public class ProcessFilters {
        ParameterExpression paramExp;
        Expression bodyFilter;
        public ProcessFilters() {
            paramExp = Expression.Parameter(typeof(Process), "p");
            bodyFilter = null;
        }

        // Create a lambda expression - we always have one single parameter
        // of type Process, that is declared as paramExp by the constructor
        public Expression<Func<Process, bool>> GetExpression() {
            if (bodyFilter == null) {
                return null;
            }
            Expression<Func<Process, bool>> filter;
            filter = Expression.Lambda<Func<Process, bool>>(
                bodyFilter,
                new ParameterExpression[] { paramExp });
            return filter;
        }

        // Add an AND with a filter on fieldName compared 
        // with comparisonValue with the required operator
        public void Add( string fieldName,
                         ExpressionType comparisonOperator, 
                         object comparisonValue) {

            switch (comparisonOperator) {
                case ExpressionType.Equal:
                case ExpressionType.NotEqual:
                case ExpressionType.LessThan:
                case ExpressionType.LessThanOrEqual:
                case ExpressionType.GreaterThan:
                case ExpressionType.GreaterThanOrEqual:
                    // Supported operations
                    break;
                default:
                    throw new NotSupportedException(
                        String.Format(
                            "Operator {0} is not supported in ProcessFilters.Add",
                            comparisonOperator.ToString()));
            } // end switch (check comparisonOperator)

            ConstantExpression comparisonConstant = 
                Expression.Constant(comparisonValue, comparisonValue.GetType());
            MemberExpression fieldAccess = 
                Expression.Property(paramExp, fieldName);
            BinaryExpression comparison = 
                Expression.MakeBinary(comparisonOperator, fieldAccess, comparisonConstant );
            if (bodyFilter == null) {
                bodyFilter = comparison;
            }
            else {
                bodyFilter = Expression.AndAlso(bodyFilter, comparison);
            }
        }
    }
    class Program {
        // 
        // Listing 11-27
        // 
        static Expression<Func<Process, bool>> FilterSelection() {
            ProcessFilters pf = new ProcessFilters();

            // Set dynamic filters
            pf.Add("Responding", ExpressionType.Equal, true );
            pf.Add("BasePriority", ExpressionType.GreaterThan, 8);

            return pf.GetExpression();
        }

        static void Main(string[] args) {
            Expression<Func<Process, bool>> filterExpression = FilterSelection();
            Display(filterExpression);
        }

        // 
        // Listing 11-26
        // 
        static void Display(Expression<Func<Process, bool>> filterExpression) {
            // If not defined, use a dummy filter
            if (filterExpression == null) {
                filterExpression = (p) => true;
            }

            Console.WriteLine("Filter : {0}", filterExpression.ToString());

            var query =
                Process.GetProcesses().AsQueryable()
                .Where(filterExpression)
                .Select(p => p.ProcessName);

            // Dump filtered  processes
            foreach (var row in query) Console.WriteLine(row);
        }
    }
}
