﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;

namespace DevLeap.Linq.ExpressionTrees.LambdaExpressions {
    class Program {
        // 
        // Listing 11-1
        // 
        delegate int IncDelegate(int x);
        static void TestDelegate() {
            IncDelegate Inc = delegate(int x) { return x + 1; };
            Console.WriteLine(Inc(2));
        }

        // 
        // Listing 11-2
        // 
        static void TestLambda() {
            IncDelegate Inc = x => x + 1;
            Console.WriteLine(Inc(2));
        }

        // 
        // Listing 11-3
        // 
        static void TestFunc() {
            // Typical lambda expression using Func delegate
            Func<int, int> IncDelegate = delegate(int x) { return x + 1; };
            Func<int, int> IncLambda = x => x + 1;

            Console.WriteLine(IncDelegate(2));
            Console.WriteLine(IncLambda(2));
        }

        //
        // Recursive Lambda Expressions box
        //
        static Func<T, T> Fix<T>(Func<Func<T, T>, Func<T, T>> F) {
            return t => F(Fix(F))(t);
        }
        static void TestRecursive() {
            Func<int, int> factorial =
                Fix<int>(
                    fac =>
                        x => (x == 0) ? 1 : x * fac(x - 1)
                        );
            Console.WriteLine(factorial(5));
        }

        // 
        // Listing 11-4
        // 
        static void DemoTriangleArea() {
            Func<double, double, double> TriangleArea = (b, h) => b * h / 2;
            Console.WriteLine("*** Delegate ***");
            Console.WriteLine("ToString: {0}", TriangleArea.ToString());
            Console.WriteLine("Value: {0}", TriangleArea(7, 12));

            Expression<Func<double, double, double>> TriangleAreaExp = 
                (b, h) => b * h / 2;
            Console.WriteLine("*** Expression tree ***");
            Console.WriteLine("ToString: {0}", TriangleAreaExp.ToString());
            Console.WriteLine("Value: {0}", TriangleAreaExp.Compile()(7, 12));
        }

        // 
        // Listing 11-5
        // 
        static int Double(int n) {
            return n * 2;
        }
        static void CallInExpression() {
            Expression<Func<int, int>> CallExp = (x) => Double(x) + 1;
            Console.WriteLine("ToString: {0}", CallExp.ToString());
        }

        // 
        // Listing 11-6
        // 
        static void DelegateInExpression() {
            Func<int,int> Double = (n) => n * 2;
            Expression<Func<int, int>> CallExp = (x) => Double(x) + 1;
            Console.WriteLine("ToString: {0}", CallExp.ToString());
        }

        // 
        // Listing 11-7
        // 
        static void ExpressionInExpression() {
            Expression<Func<int, int>> Double = (n) => n * 2;
            Expression<Func<int, int>> CallExp = (x) => Double.Compile()(x) + 1;
            Console.WriteLine("ToString: {0}", CallExp.ToString());
        }

        // 
        // Listing 11-8
        // 
        static void Immutability() {
            Expression<Func<int, int>> Formula = (n) => (n * 2 + 1) * 4;

            Console.WriteLine(Formula.ToString());

            Expression top = Formula.Body;
            ConstantExpression constant = top.Right() as ConstantExpression;
            Console.WriteLine(constant.Value);

            // constant.Value = 5; // Compiler error - Value is a read-only property
        }

        // 
        // Listing 11-9
        // 
        static void ReplaceLevel2() {
            Expression<Func<int, int>> Formula = (n) => (n * 2 + 1) * 4;

            // Replace 4 with 5
            // Results in 
            //     (n * 2 + 1) * 5
            Expression top = Formula.Body;
            ConstantExpression newRight = Expression.Constant(5);
            Expression newTree = Expression.MakeBinary(
                                     top.NodeType, 
                                     top.Left(), 
                                     newRight );
            Console.WriteLine("Original tree: {0}", top.ToString());
            Console.WriteLine("Modified tree: {0}", newTree.ToString());
        }

        // 
        // Listing 11-10
        // 
        static void ReplaceLevel3() {
            Expression<Func<int, int>> Formula = (n) => (n * 2 + 1) * 4;

            // Replace 1 with 3
            // Results in 
            //     (n * 2 + 3) * 4
            Expression top = Formula.Body;
            ConstantExpression newConstantSum = Expression.Constant(3);
            Expression newSum = Expression.MakeBinary(
                                    top.Left().NodeType, 
                                    top.Left().Left(), 
                                    newConstantSum );
            Expression newTree = Expression.MakeBinary(
                                     top.NodeType, 
                                     newSum, 
                                     top.Right() );
            Console.WriteLine("Original tree: {0}", top.ToString());
            Console.WriteLine("Modified tree: {0}", newTree.ToString());
        }

        static void ComplexTree() {
            Expression<Func<int, int>> Formula = (n) => 1 + Double(n * (n % 2 == 0 ? -1 : 1));
            Console.WriteLine(Formula.ToString());
        }

        // 
        // Listing 11-20
        // 
        static void TreeConstruction() {
            Func<int, int> lambdaInc = (n) => n + 1;
            Expression<Func<int, int>> exprInc;
            exprInc = (n) => n + 1;

            Console.WriteLine("lambdaInc : {0}", lambdaInc.ToString());
            Console.WriteLine("exprInc   : {0}", exprInc.ToString());
        }

        static void LambdaAnonymousMethod() {
            Func<int, int> lambdaInc = delegate(int n) { return n + 1; };

            Console.WriteLine("lambdaInc : {0}", lambdaInc.ToString());
        }

        // 
        // Listing 11-21
        // 
        static int IncOperation(int n) {
            return n + 1;
        }
        static void LambdaDelegate() {
            Func<int, int> lambdaInc = new Func<int,int>(IncOperation);

            Console.WriteLine("lambdaInc : {0}", lambdaInc.ToString());
        }

        // 
        // Listing 11-22
        // 
        static void ExpressionAssignment() {
            Expression<Func<int, int>> exprInc;
            ConstantExpression constant = Expression.Constant(1, typeof(int));
            ParameterExpression parameter = Expression.Parameter(typeof(int), "n");
            BinaryExpression add = Expression.Add( parameter, constant );
            exprInc = Expression.Lambda<Func<int, int>>( add, new ParameterExpression[] { parameter } );
            Console.WriteLine("exprInc   : {0}", exprInc.ToString());
        }

        // 
        // Listing 11-23
        // 
        static void TreeCombination() {
            Expression<Func<double, double, double>> rectArea = (b, h) => b * h;
            
            ParameterExpression x = Expression.Parameter(typeof(double), "x" );
            ParameterExpression y = Expression.Parameter(typeof(double), "y");
            ParameterExpression z = Expression.Parameter(typeof(double), "z");
            Expression area = Expression.Invoke(rectArea, new Expression[] { x, y });
            Expression multiply = Expression.Multiply(z, area);
            Expression<Func<double,double,double,double>> volume = 
                Expression.Lambda<Func<double,double,double,double>>(
                    multiply, 
                    new ParameterExpression[] { x, y, z });

            Console.WriteLine("Area   = {0}", rectArea.ToString());
            Console.WriteLine("Volume = {0}", volume.ToString());
            Console.WriteLine("Area value   = {0}", rectArea.Compile()(20, 10));
            Console.WriteLine("Volume value = {0}", volume.Compile()(20, 10, 8));
        }

        static void Main(string[] args) {
            //TestDelegate();
            //TestLambda();
            //TestRecursive();
            //DemoTriangleArea();
            //CallInExpression();
            //DelegateInExpression();
            //ExpressionInExpression();
            //Immutability();
            //ReplaceLevel2();
            //ReplaceLevel3();
            //ComplexTree();
            //TreeConstruction();
            //LambdaAnonymousMethod();
            //LambdaDelegate();
            //ExpressionAssignment();
            TreeCombination();
        }
    }

    // Helper extension methods to simplify the code
    // These helper methods are not safe - they throw exceptions
    // whenever applied on nodes that are not of type BinaryExpression
    public static class TreeHelper {
        public static Expression Left(this Expression exp) {
            return (exp as BinaryExpression).Left;
        }

        public static Expression Right(this Expression exp) {
            return (exp as BinaryExpression).Right;
        }

    }
}
