﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using MattWarren.ExpressionsTools;
using JomoFisher.ExpressionsTools;
using System.Collections.ObjectModel;

namespace DevLeap.Linq.ExpressionTrees.Visitor {
    // 
    // Listing 11-16
    // 
    class DisplayVisitor : ExpressionVisitor {
        private int level = 0;
        protected override Expression Visit(Expression exp) {
            if (exp != null) {
                for (int i = 0; i < level; i++) {
                    Console.Write("   ");
                }
                Console.WriteLine( "{0}  -  {1}", exp.NodeType, exp.GetType().Name );
            }
            level++;
            Expression result = base.Visit(exp);
            level--;
            return result;
        }

        public void Display(Expression exp) {
            Console.WriteLine("===== DisplayVisitor.Display =====");
            this.Visit(exp);
        }
    }

    // 
    // Listing 11-24
    // 
    class RemoveInvokeVisitor<T> : ExpressionVisitor {
        private ReadOnlyCollection<ParameterExpression> lambdaParameters;
        public RemoveInvokeVisitor(ReadOnlyCollection<ParameterExpression> parameters) {
            this.lambdaParameters = parameters;
        }
        protected override Expression VisitInvocation(InvocationExpression iv) {
            var newPars = iv.Arguments;
            LambdaExpression lambda = (iv.Expression) as LambdaExpression;
            if (lambda != null) {
                var oldPars = lambda.Parameters;
                ReplaceParametersVisitors replace =
                    new ReplaceParametersVisitors(oldPars, newPars);
                return this.Visit( replace.ReplaceVisit(lambda.Body) );
            }
            else {
                return base.VisitInvocation(iv);
            }
        }
        public Expression<T> RemoveInvokeVisit(Expression<T> exp) {
            return (Expression<T>) Visit(exp);
        }
    }

    class ReplaceParametersVisitors : ExpressionVisitor {
        private ReadOnlyCollection<Expression> newParameters;
        private ReadOnlyCollection<ParameterExpression> oldParameters;
        public ReplaceParametersVisitors(
            ReadOnlyCollection<ParameterExpression> oldParameters,
            ReadOnlyCollection<Expression> newParameters) {

            this.newParameters = newParameters;
            this.oldParameters = oldParameters;
        }
        protected override Expression VisitParameter(ParameterExpression p) {
            if (oldParameters != null
                && newParameters != null) {
                if (oldParameters.Contains(p)) {
                    return newParameters[oldParameters.IndexOf(p)];
                }
            }
            return base.VisitParameter(p);
        }
        public Expression ReplaceVisit(Expression exp) {
            return Visit(exp);
        }
    }

    class Program {
        // 
        // Listing 11-17
        // 
        static int Double(int n) {
            return n * 2;
        }
        static void DemoExpressionVisitor() {
            Expression<Func<int, int>> Formula = 
                (n) => 1 + Double(n * (n % 2 == 0 ? -1 : 1));
            Console.WriteLine(Formula.ToString());
            DisplayVisitor visitor = new DisplayVisitor();
            visitor.Display(Formula);
        }

        // 
        // Listing 11-19
        // 
        static void DemoVisitLambda() {
            Expression<Func<int, int>> Formula =
                (n) => 1 + Double(n * (n % 2 == 0 ? -1 : 1));
            Console.WriteLine(Formula.ToString());

            int level = 0;
            var visitFormula = ExprOp.Visit.Chain(
                (self, last, expr) => {
                    if (expr != null) {
                        for (int i = 0; i < level; i++) {
                            Console.Write("   ");
                        }
                        Console.WriteLine(
                            "{0}  -  {1}",
                            expr.NodeType,
                            expr.GetType().Name);
                    }
                    level++;
                    var result = last(expr);
                    level--;
                    return result;
                } 
            );

            Console.WriteLine("===== DemoVisitLambda =====");
            visitFormula(Formula);
        }

        static void TreeCombinationSubstitution() {
            Expression<Func<double, double, double>> rectArea = (b, h) => b * h;

            ParameterExpression x = Expression.Parameter(typeof(double), "x");
            ParameterExpression y = Expression.Parameter(typeof(double), "y");
            ParameterExpression z = Expression.Parameter(typeof(double), "z");
            Expression area = Expression.Invoke(rectArea, new Expression[] { x, y });
            Expression multiply = Expression.Multiply(z, area);
            Expression<Func<double, double, double, double>> volume =
                Expression.Lambda<Func<double, double, double, double>>(
                    multiply,
                    new ParameterExpression[] { x, y, z });

            Console.WriteLine("Area   = {0}", rectArea.ToString());
            Console.WriteLine("Volume = {0}", volume.ToString());
            Console.WriteLine("Area value   = {0}", rectArea.Compile()(20, 10));
            Console.WriteLine("Volume value = {0}", volume.Compile()(20, 10, 8));

            // 
            // Listing 11-25
            // 
            var cleaner = new RemoveInvokeVisitor<Func<double, double, double, double>>(volume.Parameters);
            var cleanVolume= cleaner.RemoveInvokeVisit(volume);
            Console.WriteLine("CleanVolume = {0}", cleanVolume.ToString());
            Console.WriteLine("CleanVolume value = {0}", cleanVolume.Compile()(20, 10, 8));

        }

        static void PerformanceVisitors() {
            // Short comparison of performances between the two visitors...
            Expression<Func<int, int>> Formula =
                (n) => 1 + Double(n * (n % 2 == 0 ? -1 : 1));

            int loops = 1000000;
            EmptyVisitor visitor = new EmptyVisitor();
            Stopwatch swClass = new Stopwatch();
            swClass.Start();
            for (int i = 0; i < loops; i++) {
                visitor.EmptyVisit(Formula);
            }
            swClass.Stop();
            Console.WriteLine("Class based visitor  : {0}", swClass.Elapsed);

            var visitFormula = ExprOp.Visit.Chain(
                (self, last, expr) =>
                {
                    return last(expr);
                }
            );
            Stopwatch swLambda = new Stopwatch();
            swLambda.Start();
            for (int i = 0; i < loops; i++) {
                visitFormula(Formula);
            }
            swLambda.Stop();
            Console.WriteLine("Lambda based visitor : {0}", swLambda.Elapsed);
        }


        static void TreeRecursiveSubstitution() {
            Expression<Func<double, double, double>> rectArea = (b, h) => b * h;

            ParameterExpression x = Expression.Parameter(typeof(double), "x");
            ParameterExpression y = Expression.Parameter(typeof(double), "y");
            ParameterExpression z = Expression.Parameter(typeof(double), "z");
            Expression area = Expression.Invoke(rectArea, new Expression[] { x, y });
            Expression multiply = Expression.Multiply(z, area);
            Expression<Func<double, double, double, double>> volume =
                Expression.Lambda<Func<double, double, double, double>>(
                    multiply,
                    new ParameterExpression[] { x, y, z });

            ConstantExpression p1 = Expression.Constant(10.0, typeof(double));
            ConstantExpression p2 = Expression.Constant(20.0, typeof(double));
            ParameterExpression m = Expression.Parameter(typeof(double), "m");

            Expression hyper = Expression.Invoke(volume, new Expression[] { p1, p2, m});
            Expression multiply2 = Expression.Multiply(m, hyper);
            Expression<Func<double, double>> v2 =
                Expression.Lambda<Func<double, double>>(
                    multiply2,
                    new ParameterExpression[] { m });


            Console.WriteLine("Area   = {0}", rectArea.ToString());
            Console.WriteLine("Volume = {0}", volume.ToString());
            Console.WriteLine("V2     = {0}", v2.ToString());
            Console.WriteLine("Area value   = {0}", rectArea.Compile()(20, 10));
            Console.WriteLine("Volume value = {0}", volume.Compile()(20, 10, 8));
            Console.WriteLine("V2     value = {0}", v2.Compile()(20));

            // 
            // Listing 11-25
            // 
            var cleaner = new RemoveInvokeVisitor<Func<double, double, double, double>>(volume.Parameters);
            var cleanVolume = cleaner.RemoveInvokeVisit(volume);
            Console.WriteLine("CleanVolume = {0}", cleanVolume.ToString());
            Console.WriteLine("CleanVolume value = {0}", cleanVolume.Compile()(20, 10, 8));

            var cleaner2 = new RemoveInvokeVisitor<Func<double, double>>(v2.Parameters);
            var cleanHyper = cleaner2.RemoveInvokeVisit(v2);
            Console.WriteLine("CleanVolume = {0}", cleanHyper.ToString());
            Console.WriteLine("CleanVolume value = {0}", cleanHyper.Compile()(9));

        }


        static void Main(string[] args) {
            DemoExpressionVisitor();
            //DemoVisitLambda();
            //TreeCombinationSubstitution();
            //TreeRecursiveSubstitution();
            // PerformanceVisitors();
        }
    }
    class EmptyVisitor : ExpressionVisitor {
        protected override Expression Visit(Expression exp) {
            return base.Visit(exp);
        }

        public void EmptyVisit(Expression exp) {
            this.Visit(exp);
        }
    }

}
