﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DevLeap.Linq.ExtendingLINQ.CustomOperator {
    static class Extensions {
        // 
        // Listing 12-6
        // 
        // Generic version of the MinItem operator
        public static TSource MinItem<TSource, TCompareValue>(
                this IEnumerable<TSource> source,
                Func<TSource, TCompareValue> comparerExpression) {

            // We get a constructed Comparer<T> instance 
            // for the type to be compared
            Comparer<TCompareValue> comparer = Comparer<TCompareValue>.Default;
            // indexElement is used to execute two different pieces of code
            // in the lambda expression for Aggregate: one for the very 
            // first iteration, the other for the following iterations
            int indexElement = 0;
            // The Aggregate will execute the lambda expression 
            // for each element in the source sequence
            return source.Aggregate(
                default(TSource),
                (minValue, item) =>
                       (indexElement++ == 0 ?
                            // First iteration - no comparison, 
                            // returns item as initial minValue
                            item :
                            // Secondo or more iteration: do the comparison
                            // and returns item if it is lower then the 
                            // previous minValue, otherwise returns minValue
                            (comparer.Compare(comparerExpression(item),
                                               comparerExpression(minValue)) < 0 ?
                                item :
                                minValue)));
        }

        // 
        // Listing 12-4
        // 
        // Specific version for Quote of the MinItem operator
        public static Quote MinPrice( this IEnumerable<Quote> source) {
            return source.Aggregate( (t, s) => t.Price < s.Price ? t : s );
        }

        // 
        // Listing 12-13
        // 
        // Specialization of the Min operator for Quote instances
        // WARNING: this is a dangerous behavior, because in this case we are changing the semantic
        //          of the Min operator, which does not return the compared expression anymore
        public static TSource Min<TSource, TCompareValue>(
                this IEnumerable<TSource> source, 
                Func<TSource, TCompareValue> comparerExpression)
            where TSource : Quote {

            return source.MinItem(comparerExpression);
        }
    }

    class Program {
        static void Main(string[] args) {
            // Use en-US setting to get the same result as the printed book!
            System.Threading.Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-US", true);

            // 
            // Listing 12-2
            // 
            Stock stock = new Stock { Name = "Stock Demo", Ticker = "AGHQ", Shares = 100000M };
            stock.Quotes = new List<Quote> { 
                new Quote { Stock = stock, Time = new DateTime(2007, 12, 5, 11, 8, 45 ), Price = 57.63M },
                new Quote { Stock = stock, Time = new DateTime(2007, 12, 5, 11, 8, 56 ), Price = 56.92M },
                new Quote { Stock = stock, Time = new DateTime(2007, 12, 5, 11, 9, 08 ), Price = 57.05M },
                new Quote { Stock = stock, Time = new DateTime(2007, 12, 5, 11, 9, 23 ), Price = 56.87M } 
            };

            // 
            // Listing 12-3
            // 
            Console.WriteLine("Stock: {0}", stock);

            Quote minQuote;

            // Get the item with the minimum price using Aggregate operator
            minQuote = stock.Quotes.Aggregate(
                           (t, s) => t.Price < s.Price ? t : s );
            Console.WriteLine(
                "Min item using Aggregate - {0} : {1}", 
                minQuote.Time, 
                minQuote.Price);

            // 
            // Listing 12-5
            // 
            // Get the item with the minimum price using a specific custom operator
            minQuote = stock.Quotes.MinPrice();
            Console.WriteLine(
                "Min item using MinPrice  - {0} : {1}",
                minQuote.Time,
                minQuote.Price);

            // 
            // Listing 12-7
            // 
            // Get the item with the minimum price using a generic custom operator
            minQuote = stock.Quotes.MinItem(x => x.Price);
            Console.WriteLine(
                "Min item using MinItem(price) - {0} : {1}",
                minQuote.Time,
                minQuote.Price);

            // Get the item with the minimum time using a generic custom operator
            minQuote = stock.Quotes.MinItem(x => x.Time);
            Console.WriteLine(
                "Min item using MinItem(time)  - {0} : {1}",
                minQuote.Time,
                minQuote.Price);

            // 
            // Listing 12-14
            // 
            // Get the item with the minimum price using a generic custom operator
            minQuote = stock.Quotes.Min(x => x.Price);
            Console.WriteLine(
                "Min item using Min(price) - {0} : {1}",
                minQuote.Time,
                minQuote.Price);

            // Get the item with the minimum time using a generic custom operator
            minQuote = stock.Quotes.Min(x => x.Time);
            Console.WriteLine(
                "Min item using Min(time)  - {0} : {1}",
                minQuote.Time,
                minQuote.Price);
        }
    }
}
