﻿// 
// The code in this listing has been written Jomo Fisher on his blog:
/// - http://blogs.msdn.com/jomo_fisher/archive/2007/05/23/dealing-with-linq-s-immutable-expression-trees.aspx
// and has been released under the Microsoft Public License published below
// 

/*
Microsoft Public License (Ms-PL)
Published: October 12, 2006
  
Related Links
• Shared Source Licenses 
• Microsoft Limited Public License 
• Microsoft Reciprocal License 
• Microsoft Reference License 
  
 
This license governs use of the accompanying software. If you use the software, you accept this license. If you do not accept the license, do not use the software.

1. Definitions
The terms “reproduce,” “reproduction,” “derivative works,” and “distribution” have the same meaning here as under U.S. copyright law.

A “contribution” is the original software, or any additions or changes to the software.

A “contributor” is any person that distributes its contribution under this license.

“Licensed patents” are a contributor’s patent claims that read directly on its contribution.

Top of page
2. Grant of Rights
(A) Copyright Grant- Subject to the terms of this license, including the license conditions and limitations in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free copyright license to reproduce its contribution, prepare derivative works of its contribution, and distribute its contribution or any derivative works that you create.

(B) Patent Grant- Subject to the terms of this license, including the license conditions and limitations in section 3, each contributor grants you a non-exclusive, worldwide, royalty-free license under its licensed patents to make, have made, use, sell, offer for sale, import, and/or otherwise dispose of its contribution in the software or derivative works of the contribution in the software.

Top of page
3. Conditions and Limitations
(A) No Trademark License- This license does not grant you rights to use any contributors’ name, logo, or trademarks.

(B) If you bring a patent claim against any contributor over patents that you claim are infringed by the software, your patent license from such contributor to the software ends automatically.

(C) If you distribute any portion of the software, you must retain all copyright, patent, trademark, and attribution notices that are present in the software.

(D) If you distribute any portion of the software in source code form, you may do so only under this license by including a complete copy of this license with your distribution. If you distribute any portion of the software in compiled or object code form, you may only do so under a license that complies with this license.

(E) The software is licensed “as-is.” You bear the risk of using it. The contributors give no express warranties, guarantees or conditions. You may have additional consumer rights under your local laws which this license cannot change. To the extent permitted under your local laws, the contributors exclude the implied warranties of merchantability, fitness for a particular purpose and non-infringement.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Collections.ObjectModel;

namespace DevLeap.Linq.ExtendingLINQ.PredicateAnalysis {
    /// <summary>
    /// This class was originally published by Jomo Fisher on his blog:
    /// - http://blogs.msdn.com/jomo_fisher/archive/2007/05/23/dealing-with-linq-s-immutable-expression-trees.aspx
    /// </summary>
    public static class ExprOp {
        static public Func<Expression, Expression> Visit = 
            FuncOp.Create<Expression, Expression>(
                (self, expr) => {
                if (expr == null) return expr;  // Line added by Marco Russo - it was not in the original code
                                                // it is necessary to handle null values in visited child nodes
                                                // like, for example, the Object of a MethodCallExpression on a
                                                // static method
                switch (expr.NodeType) {
                    case ExpressionType.Coalesce:
                        var c = (BinaryExpression) expr;
                        var left = self(c.Left);
                        var right = self(c.Right);
                        var conv = self(c.Conversion);
                        return (left == c.Left 
                                 && right == c.Right 
                                 && conv == c.Conversion) 
                                ? expr 
                                : Expression.Coalesce(
                                      left, 
                                      right, 
                                      (LambdaExpression) conv);
                    case ExpressionType.Conditional:
                        var ce = (ConditionalExpression) expr;
                        var t = self(ce.Test);
                        var it = self(ce.IfTrue);
                        var @if = self(ce.IfFalse);
                        return (t == ce.Test 
                                 && it == ce.IfTrue 
                                 && @if == ce.IfFalse) 
                               ? expr 
                               : Expression.Condition(t, it, @if);
                    case ExpressionType.TypeIs:
                        var tbe = (TypeBinaryExpression) expr;
                        var tbex = self(tbe.Expression);
                        return (tbe.Expression == tbex) ? expr : Expression.TypeIs(tbex, tbe.TypeOperand);
                    case ExpressionType.MemberAccess:
                        var ma = (MemberExpression) expr;
                        var maex = self(ma.Expression);
                        return (maex == ma.Expression) ? expr : Expression.MakeMemberAccess(maex, ma.Member);
                    case ExpressionType.Call:
                        var mce = (MethodCallExpression) expr;
                        var o = self(mce.Object);
                        var ca = self.VisitExpressionList(mce.Arguments);
                        return (o == mce.Object && ca == mce.Arguments) ? expr : Expression.Call(o, mce.Method, ca);
                    case ExpressionType.Lambda:
                        var le = (LambdaExpression) expr;
                        var b = self(le.Body);
                        return (b == le.Body) ? expr : Expression.Lambda(le.Type, b, le.Parameters);
                    case ExpressionType.New:
                        var ne = (NewExpression) expr;
                        var nar = self.VisitExpressionList(ne.Arguments);
                        if (nar != ne.Arguments)
                            return (ne.Members != null) ? Expression.New(ne.Constructor, nar, ne.Members)
                                : Expression.New(ne.Constructor, nar);
                        return expr;
                    case ExpressionType.NewArrayInit:
                    case ExpressionType.NewArrayBounds:
                        var na = (NewArrayExpression) expr;
                        var inits = self.VisitExpressionList(na.Expressions);
                        if (inits != na.Expressions)
                            return (na.NodeType == ExpressionType.NewArrayInit) ?
                                Expression.NewArrayInit(na.Type.GetElementType(), inits) :
                                Expression.NewArrayBounds(na.Type.GetElementType(), inits);
                        return expr;
                    case ExpressionType.Invoke:
                        var inv = (InvocationExpression) expr;
                        var args = self.VisitExpressionList(inv.Arguments);
                        var ie = self(inv.Expression);
                        return (args == inv.Arguments && ie == inv.Expression) ? expr : Expression.Invoke(ie, args);
                    case ExpressionType.MemberInit:
                        var mi = (MemberInitExpression) expr;
                        var n = (NewExpression) self(mi.NewExpression);
                        var bindings = self.VisitBindingList(mi.Bindings);
                        return (n == mi.NewExpression && bindings == mi.Bindings) ? expr : Expression.MemberInit(n, bindings);
                    case ExpressionType.ListInit:
                        var li = (ListInitExpression) expr;
                        var lin = (NewExpression) self(li.NewExpression);
                        var lii = VisitElementInitializerList(self, li.Initializers);
                        return (lin == li.NewExpression || lii == li.Initializers) ? expr : Expression.ListInit(lin, lii);
                }
                if (expr.IsBinary()) {
                    var b = (BinaryExpression) expr;
                    var left = self(b.Left);
                    var right = self(b.Right);
                    return (left == b.Left && right == b.Right) 
                           ? expr 
                           : Expression.MakeBinary(
                                  expr.NodeType, 
                                  left, 
                                  right);
                }
                else if (expr.IsUnary()) {
                    var u = (UnaryExpression) expr;
                    var op = self(u.Operand);
                    return (u.Operand == op) 
                           ? expr 
                           : Expression.MakeUnary(
                                  u.NodeType, 
                                  op, 
                                  expr.Type);
                }
                return expr;
            }
        );
        public static bool IsBinary(this Expression expr) {
            return expr is BinaryExpression;
        }
        public static bool IsUnary(this Expression expr) {
            return expr is UnaryExpression;
        }
        public static MemberBinding VisitBinding(this Func<Expression, Expression> self, MemberBinding b) {
            switch (b.BindingType) {
                case MemberBindingType.Assignment:
                    return self.VisitMemberAssignment((MemberAssignment) b);
                case MemberBindingType.MemberBinding: return self.VisitMemberMemberBinding((MemberMemberBinding) b);
            }
            return self.VisitMemberListBinding((MemberListBinding) b);
        }
        public static MemberAssignment VisitMemberAssignment(this Func<Expression, Expression> self, MemberAssignment assignment) {
            var e = self(assignment.Expression);
            return (e == assignment.Expression) ? assignment : Expression.Bind(assignment.Member, e);
        }
        public static MemberMemberBinding VisitMemberMemberBinding(this Func<Expression, Expression> self, MemberMemberBinding binding) {
            var bindings = self.VisitBindingList(binding.Bindings);
            return (bindings == binding.Bindings) ? binding : Expression.MemberBind(binding.Member, bindings);
        }
        public static MemberListBinding VisitMemberListBinding(this Func<Expression, Expression> self, MemberListBinding binding) {
            var initializers = self.VisitElementInitializerList(binding.Initializers);
            return (initializers == binding.Initializers) ? binding : Expression.ListBind(binding.Member, initializers);
        }
        public static ElementInit VisitElementInitializer(this Func<Expression, Expression> self, ElementInit initializer) {
            var arguments = self.VisitExpressionList(initializer.Arguments);
            return (arguments == initializer.Arguments) ? initializer : Expression.ElementInit(initializer.AddMethod, arguments);
        }
        public static ReadOnlyCollection<Expression> VisitExpressionList(this Func<Expression, Expression> self, ReadOnlyCollection<Expression> original) {
            return VisitList(original, e => self(e));
        }
        public static ReadOnlyCollection<MemberBinding> VisitBindingList(this Func<Expression, Expression> self, ReadOnlyCollection<MemberBinding> original) {
            return VisitList(original, e => self.VisitBinding(e));
        }
        public static ReadOnlyCollection<ElementInit> VisitElementInitializerList(this Func<Expression, Expression> self, ReadOnlyCollection<ElementInit> original) {
            return VisitList(original, e => VisitElementInitializer(self, e));
        }
        private static ReadOnlyCollection<T> VisitList<T>(ReadOnlyCollection<T> original, Func<T, T> op) {
            List<T> @new = null;
            for (int i = 0, n = original.Count; i < n; i++) {
                T init = op(original[i]);
                if (@new != null)
                    @new.Add(init);
                else if (!ReferenceEquals(init, original[i])) {
                    @new = new List<T>(n);
                    for (int j = 0; j < i; j++)
                        @new.Add(original[j]);
                    @new.Add(init);
                }
            }
            return (@new == null) ? original : @new.AsReadOnly();
        }
    }

}
