﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Diagnostics;
using System.Text;

namespace DevLeap.Linq.ExtendingLINQ.PredicateAnalysis {
    static class Extension {
        // 
        // Listing 12-19
        // 
        /// <summary>
        /// Limits:
        ///   - TKey must be a reference type
        ///   - If there are multiple statements as "Key == [constant]", only the last one visited is considered
        ///   - There is no check if the Key comparison is made inside a more complex logical condition
        ///     like an OR - we assume that it is in AND with everything else
        ///   - only the string class used as TKey has been tested (TKey must be a reference type)
        /// </summary>
        public static IEnumerable<KeyValuePair<TKey, TValue>> Where<TKey, TValue>(
                this SortedDictionary<TKey, TValue> source, 
                Expression<Func<KeyValuePair<TKey, TValue>, Boolean>> predicate)
                    where TKey : class {
            // This is for trace only: we display the predicate we've found
            Trace.WriteLine("** Where predicate **");
            Trace.WriteLine(predicate.ToString());
            Trace.WriteLine("** --------------- **");

            // After visiting the expression tree for the predicate, we need to 
            // know if a standard or an optimized Where has to be done
            bool replaceWithWhereKey = false;
            TKey keyToSearch = null;

            #region Key comparison replacement
            // The following expression results in a delegate that returns a copy
            // of the expression tree, eliminating an expression like
            //   <item>.Key == <constant>
            // where <item> is an instance of KeyValuePair<TKey,TValue> (an element
            // of our SortedDictionary) and <constant> is a constant value (we tested
            // only strings)
            var ReplaceEqual = ExprOp.Visit.Chain(
                (self, last, expr) =>
                {
                    if (expr == null) return null;
                    // Visiting the expression tree, stop at a BinaryExpression 
                    // with an Equal comparison
                    switch (expr.NodeType) {
                        case ExpressionType.Equal:
                            var b = (BinaryExpression) expr;
                            TKey _key = null;
                            MemberExpression memberAccess = null;
                            // Get MemberAccess and Constant regardless of their position
                            switch (b.Left.NodeType) {
                                case ExpressionType.MemberAccess:
                                    memberAccess = b.Left as MemberExpression;
                                    break;
                                case ExpressionType.Constant:
                                    _key = (b.Left as ConstantExpression).Value as TKey;
                                    break;
                            }
                            switch (b.Right.NodeType) {
                                case ExpressionType.MemberAccess:
                                    memberAccess = b.Right as MemberExpression;
                                    break;
                                case ExpressionType.Constant:
                                    _key = (b.Right as ConstantExpression).Value as TKey;
                                    break;
                            }
                            // Stops here without modifications if either
                            // memberAccess or _key have not been found
                            if ((memberAccess == null) || (_key == null)) {
                                return b;
                            }
                            // If we access to the Key property of a 
                            // KeyValuePair<TKey,TValue> type, we can do 
                            // the substitution - our BinaryExpression is 
                            // replaced with a constant value equals to true
                            if ((memberAccess.Member.ReflectedType == typeof(KeyValuePair<TKey, TValue>))
                                 && (memberAccess.Member.Name == "Key")) {
                                // Set flag to replace the where condition 
                                // in the caller (the Where operator)
                                replaceWithWhereKey = true;
                                keyToSearch = _key;
                                return Expression.Constant(true);
                            }
                            Console.WriteLine(b.Left.ToString());
                            return b;
                        default:
                            return last(expr);
                    } // end switch (expr.NodeType)
                } // end anonymous delegate
            );
            #endregion

            // The following call eliminates an 
            //   <item>.Key == <constant>
            // comparison and makes a direct access to the Dictionary
            // using the <constant> found (saved in keyToSearch) if that
            // comparison has been found and removed
            var querySubstitution = ReplaceEqual(predicate);
            if (replaceWithWhereKey) {
                Trace.WriteLine("--- REPLACED WHERE ---");
                // Directly get the value associated towith the given key
                // thanks to the SortedDictionary implementation
                var value = source[keyToSearch];
                var item = new KeyValuePair<TKey, TValue>(keyToSearch, value);
                if (predicate.Compile()(item)) {
                    yield return item;
                }
            }
            else {
                // The traditional Where iteration takes place
                // if an optimized access has not been requested
                Trace.WriteLine("--- STANDARD WHERE ---");
                foreach (var item in source) {
                    if (predicate.Compile()(item)) {
                        yield return item;
                    }
                }
            }
        }
    }

    class Program {
        static void Main(string[] args) {
            DemoOptimization();
        }

        //
        // Listing 12-20
        //
        static void DemoOptimization() {
            Trace.Listeners.Add(new TextWriterTraceListener(Console.Out));
            SortedDictionary<string, string> list = new SortedDictionary<string, string>();
            for (char ch = 'Z'; ch >= 'A'; ch--) {
                list.Add(ch.ToString(), "Letter " + ch.ToString());
            }

            //Console.WriteLine("----------");
            //Dump(
            //    from   l in list 
            //    select l.Value);

            Console.WriteLine("-- queryOptimized --");
            var queryOptimized =
                from l in list
                where l.Key == "M" && l.Value.Length > 1
                select l.Value;
            Dump(queryOptimized);
            Console.WriteLine("");

            Console.WriteLine("-- queryNotOptimized --");
            var queryNotOptimized =
                from l in list
                where l.Value == "Letter M" && l.Value.Length > 1
                select l.Value;
            Dump(queryNotOptimized);
            Console.WriteLine("");
        }
        static void Dump<T>(IEnumerable<T> sequence) {
            foreach (T item in sequence) {
                Console.WriteLine(item);
            }
        }
    }
}
