﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DevLeap.Linq.ExtendingLINQ.SortedList {
    static class Extensions {

        //
        // Listing 12-16
        //
        public static IEnumerable<TSource> WhereKey<TSource,TKey>(
                this  SortedDictionary<TKey, TSource> source,
                TKey key) {
            
            yield return source[key];
            
            // 
            // To avoid an exception you could use ContainsKey before
            // but this would double the number of comparisons
            // 
            //if (source.ContainsKey(key))
            //    yield return source[key];
        }

        // 
        // Special WhereKey implementation: using KeyWrapper
        // we intercept all accesses to the Value properties and 
        // we can display comparisons that are done iterating
        // collection or finding a key in the key tree of the
        // sorted collection
        // 
        public static IEnumerable<TSource> WhereKey<TSource>(
                this  SortedDictionary<KeyWrapper, TSource> source,
                string key) {

            yield return source[new KeyWrapper(key)];

            // 
            // To avoid an exception you could use ContainsKey before
            // but this would double the number of comparisons
            // 
            //KeyWrapper keyWrapper = new KeyWrapper(key);
            //if (source.ContainsKey(keyWrapper))
            //    yield return source[keyWrapper];
        }

        // 
        // Possible standard Where implementation
        // 
        //public static IEnumerable<TSource> Where<TSource>(
        //        this IEnumerable<TSource> source,
        //        Func<TSource, Boolean> predicate) {
        //    foreach (TSource item in source) {
        //        if (predicate(item)) {
        //            yield return item;
        //        }
        //    }
        //}


        public static IEnumerable<KeyValuePair<TKey, TSource>> Where<TKey, TSource>(
                this SortedDictionary<TKey, TSource> source,
                TKey key ) {
            IEnumerator<KeyValuePair<TKey,TSource>> fastScan = source.MoveFirstKey( key );
            if (fastScan.Current.Key != null) {
                while (fastScan.Current.Key.Equals(key)) {
                    yield return fastScan.Current;
                    if (!fastScan.MoveNext()) break;
                }
            }
        }

        /// <summary>
        /// The MoveFirstKey is purely illustrative, it does not optimize the access using the Key btree but it iterates Keys.
        /// THe scope of this operator is only to illustrate the syntax of a Where specialization
        /// </summary>
        public static IEnumerator<KeyValuePair<TKey, TSource>> MoveFirstKey<TKey, TSource>(this SortedDictionary<TKey, TSource> dictionary, TKey key) {
            var scan = dictionary.AsEnumerable().GetEnumerator();
            while (scan.MoveNext()) {
                if (scan.Current.Key.Equals(key)) {
                    return scan;
                }
            }
            return scan;
        }
    }


    class Program {
        static void Main(string[] args) {
            //DemoSortedList();
            DemoSortedListTrace();
            DemoSemantic();
        }
        // 
        // Listing 12-15
        // 
        static void DemoSortedList() {
            SortedDictionary<string, string> list = new SortedDictionary<string, string>();
            for (char ch = 'Z'; ch >= 'A'; ch--) {
                list.Add(ch.ToString(), "Letter " + ch.ToString());
            }

            //Console.WriteLine("----------");
            //Dump(
            //    from   l in list 
            //    select l.Value);

            Console.WriteLine("-- queryStandard --");
            var queryStandard = 
                from   l in list 
                where  l.Key == "M" 
                select l.Value;
            Dump(queryStandard);

            Console.WriteLine("-- queryFast --");
            var queryFast = list.WhereKey("M");
            Dump(queryFast);

            Console.WriteLine("-- queryFastProjection --");
            var queryFastProjection =
                from l in list.WhereKey("M")
                select new { Value = l, Length = l.Length };
            Dump(queryFastProjection);
        }
        static void Dump<T>(IEnumerable<T> sequence) {
            foreach (T item in sequence) {
                Console.WriteLine(item);
            }
        }

        // 
        // Listing 12-18
        // 
        /// <summary>
        /// This function uses the KeyWrapper version, which 
        /// </summary>
        static void DemoSortedListTrace() {
            SortedDictionary<KeyWrapper, string> list = new SortedDictionary<KeyWrapper, string>();

            for (char ch = 'Z'; ch >= 'A'; ch--) {
                list.Add(new KeyWrapper(ch.ToString()), "Letter " + ch.ToString());
            }

            //Console.WriteLine("----------");
            //Dump(
            //    from l in list
            //    select l.Value);

            Console.WriteLine();
            Console.WriteLine("-- queryStandard --");
            var queryStandard =
                from l in list
                where l.Key.Value == "M"
                select l.Value;
            Dump(queryStandard);

            Console.WriteLine();
            Console.WriteLine("-- queryFast --");
            var queryFast = list.WhereKey("M");
            Dump(queryFast);
        }
        static void DemoSemantic() {
            SortedDictionary<string, string> list = new SortedDictionary<string, string>();
            for (char ch = 'Z'; ch >= 'A'; ch--) {
                list.Add(ch.ToString(), "Letter " + ch.ToString());
            }

            Console.WriteLine("-- querySemantic --");
            var querySemantic =
                from l in list.Where("M")
                select l.Value;
            Dump(querySemantic);

        }

    }
}
